import { __awaiter } from "tslib";
import * as anchor from "@project-serum/anchor";
import * as web3 from "@solana/web3.js";
import { CLAIM_REQUEST_SEED, ENTRY_SEED, NAMESPACE_SEED, NAMESPACES_IDL, NAMESPACES_PROGRAM_ID, REVERSE_ENTRY_SEED, } from ".";
export function getNamespaceByName(connection, namespaceName) {
    return __awaiter(this, void 0, void 0, function* () {
        const [namespaceId] = yield web3.PublicKey.findProgramAddress([
            anchor.utils.bytes.utf8.encode(NAMESPACE_SEED),
            anchor.utils.bytes.utf8.encode(namespaceName),
        ], NAMESPACES_PROGRAM_ID);
        return getNamespace(connection, namespaceId);
    });
}
export function getNamespace(connection, namespaceId) {
    return __awaiter(this, void 0, void 0, function* () {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        const provider = new anchor.Provider(connection, null, {});
        const namespacesProgram = new anchor.Program(NAMESPACES_IDL, NAMESPACES_PROGRAM_ID, provider);
        const parsed = (yield namespacesProgram.account.namespace.fetch(namespaceId));
        return {
            parsed,
            pubkey: namespaceId,
        };
    });
}
export function getNamespaces(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-call, @typescript-eslint/no-unsafe-assignment
        const coder = new anchor.Coder(NAMESPACES_IDL);
        const programAccounts = yield connection.getProgramAccounts(NAMESPACES_PROGRAM_ID);
        const namespaces = [];
        programAccounts.forEach((account) => {
            try {
                // eslint-disable-next-line @typescript-eslint/no-unsafe-call, @typescript-eslint/no-unsafe-member-access
                const namespace = coder.accounts.decode("namespace", account.account.data);
                namespaces.push(Object.assign(Object.assign({}, account), { parsed: namespace }));
            }
            catch (e) {
                console.log(`Failed to decode namespace`);
            }
        });
        return namespaces.sort((a, b) => a.pubkey.toBase58().localeCompare(b.pubkey.toBase58()));
    });
}
export function getNameEntry(connection, namespaceName, entryName) {
    return __awaiter(this, void 0, void 0, function* () {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        const provider = new anchor.Provider(connection, null, {});
        const namespacesProgram = new anchor.Program(NAMESPACES_IDL, NAMESPACES_PROGRAM_ID, provider);
        const [namespaceId] = yield web3.PublicKey.findProgramAddress([
            anchor.utils.bytes.utf8.encode(NAMESPACE_SEED),
            anchor.utils.bytes.utf8.encode(namespaceName),
        ], namespacesProgram.programId);
        const [entryId] = yield web3.PublicKey.findProgramAddress([
            anchor.utils.bytes.utf8.encode(ENTRY_SEED),
            namespaceId.toBytes(),
            anchor.utils.bytes.utf8.encode(entryName),
        ], namespacesProgram.programId);
        const parsed = (yield namespacesProgram.account.entry.fetch(entryId));
        return {
            parsed,
            pubkey: entryId,
        };
    });
}
export function getNameEntries(connection, namespaceName, entryNames) {
    return __awaiter(this, void 0, void 0, function* () {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        const provider = new anchor.Provider(connection, null, {});
        const namespacesProgram = new anchor.Program(NAMESPACES_IDL, NAMESPACES_PROGRAM_ID, provider);
        const [namespaceId] = yield web3.PublicKey.findProgramAddress([
            anchor.utils.bytes.utf8.encode(NAMESPACE_SEED),
            anchor.utils.bytes.utf8.encode(namespaceName),
        ], namespacesProgram.programId);
        const entryTuples = yield Promise.all(entryNames.map((entryName) => web3.PublicKey.findProgramAddress([
            anchor.utils.bytes.utf8.encode(ENTRY_SEED),
            namespaceId.toBytes(),
            anchor.utils.bytes.utf8.encode(entryName),
        ], namespacesProgram.programId)));
        const entryIds = entryTuples.map((tuple) => tuple[0]);
        const result = (yield namespacesProgram.account.entry.fetchMultiple(entryIds));
        return result.map((parsed, i) => ({
            parsed,
            pubkey: entryIds[i],
            name: entryNames[i],
        }));
    });
}
export function getClaimRequest(connection, namespaceName, entryName, requestor) {
    return __awaiter(this, void 0, void 0, function* () {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        const provider = new anchor.Provider(connection, null, {});
        const namespacesProgram = new anchor.Program(NAMESPACES_IDL, NAMESPACES_PROGRAM_ID, provider);
        const [namespaceId] = yield web3.PublicKey.findProgramAddress([
            anchor.utils.bytes.utf8.encode(NAMESPACE_SEED),
            anchor.utils.bytes.utf8.encode(namespaceName),
        ], namespacesProgram.programId);
        const [claimRequestId] = yield web3.PublicKey.findProgramAddress([
            anchor.utils.bytes.utf8.encode(CLAIM_REQUEST_SEED),
            namespaceId.toBytes(),
            anchor.utils.bytes.utf8.encode(entryName),
            requestor.toBytes(),
        ], namespacesProgram.programId);
        const parsed = (yield namespacesProgram.account.claimRequest.fetch(claimRequestId));
        return {
            parsed,
            pubkey: claimRequestId,
        };
    });
}
export function getPendingClaimRequests(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-call
        const coder = new anchor.Coder(NAMESPACES_IDL);
        const programAccounts = yield connection.getProgramAccounts(NAMESPACES_PROGRAM_ID);
        const pendingClaimRequests = [];
        programAccounts.forEach((account) => {
            try {
                // eslint-disable-next-line @typescript-eslint/no-unsafe-call, @typescript-eslint/no-unsafe-member-access
                const claimRequest = coder.accounts.decode("claimRequest", account.account.data);
                if (!claimRequest.isApproved) {
                    pendingClaimRequests.push(Object.assign(Object.assign({}, account), { parsed: claimRequest }));
                }
            }
            catch (e) {
                console.log(`Failed to decode claim request`);
            }
        });
        return pendingClaimRequests;
    });
}
export function getReverseEntry(connection, pubkey) {
    return __awaiter(this, void 0, void 0, function* () {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        const provider = new anchor.Provider(connection, null, {});
        const namespacesProgram = new anchor.Program(NAMESPACES_IDL, NAMESPACES_PROGRAM_ID, provider);
        const [reverseEntryId] = yield web3.PublicKey.findProgramAddress([anchor.utils.bytes.utf8.encode(REVERSE_ENTRY_SEED), pubkey.toBytes()], namespacesProgram.programId);
        const parsed = (yield namespacesProgram.account.reverseEntry.fetch(reverseEntryId));
        return {
            parsed,
            pubkey: reverseEntryId,
        };
    });
}
export function tryGetReverseEntry(connection, pubkey) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            return yield getReverseEntry(connection, pubkey);
        }
        catch (e) {
            console.log(`Failed to get reverse entry`);
            return null;
        }
    });
}
//# sourceMappingURL=accounts.js.map